# -*- coding: utf-8 -*-

import time
import logging
from logging import StreamHandler, Formatter
from logging.handlers import RotatingFileHandler


from cloud4rpi.config import mqqtBrokerHost
from cloud4rpi.config import mqttBrokerPort, mqttsBrokerPort
from cloud4rpi.config import loggerName

from cloud4rpi.device import Device
from cloud4rpi.mqtt_api import MqttApi
from cloud4rpi.errors import get_error_message

log = logging.getLogger(loggerName)
log.setLevel(logging.INFO)
log.addHandler(StreamHandler())


def connect(device_token,
            host=mqqtBrokerHost,
            port=None,
            tls_config=None):
    if port is None:
        port = mqttsBrokerPort if isinstance(tls_config, dict) \
            else mqttBrokerPort
    api = MqttApi(device_token, host, port, tls_config)
    __attempt_to_connect_with_retries(api)
    return Device(api)


def __attempt_to_connect_with_retries(api, attempts=10):
    retry_interval = 5
    for attempt in range(attempts):
        try:
            api.connect()
        except Exception as e:
            log.debug('MQTT connection error %s. Attempt %s', e, attempt)
            time.sleep(retry_interval)
            continue
        else:
            break
    else:

        raise Exception('Impossible to connect to MQTT broker. Quiting.')


def set_logging_to_file(log_file_path):
    log_file = RotatingFileHandler(
        log_file_path,
        maxBytes=1024 * 1024,
        backupCount=10
    )
    log_file.setFormatter(Formatter('%(asctime)s: %(message)s'))
    log.addHandler(log_file)


def set_logging_level(level=logging.INFO):
    log.setLevel(level)



# -*- coding: utf-8 -*-

import sys
import re
import inspect
import logging
import numbers
from math import isnan, isinf
from datetime import datetime, tzinfo, timedelta
from cloud4rpi import config
from cloud4rpi.errors import InvalidTokenError
from cloud4rpi.errors import InvalidConfigError
from cloud4rpi.errors import UnexpectedVariableTypeError
from cloud4rpi.errors import UnexpectedVariableValueTypeError
from cloud4rpi.errors import TYPE_WARN_MSG

if sys.version_info[0] > 2:
    from cloud4rpi.utils_v3 import is_string
else:
    from cloud4rpi.utils_v2 import is_string

log = logging.getLogger(config.loggerName)

BOOL_TYPE = 'bool'
NUMERIC_TYPE = 'numeric'
STRING_TYPE = 'string'
LOCATION_TYPE = 'location'

SUPPORTED_VARIABLE_TYPES = [BOOL_TYPE, NUMERIC_TYPE, STRING_TYPE, LOCATION_TYPE]


class UtcTzInfo(tzinfo):
    # pylint: disable=W0223
    def tzname(self, dt):
        return "UTC"

    def utcoffset(self, dt):
        return timedelta(0)


def guard_against_invalid_token(token):
    token_re = re.compile('[1-9a-km-zA-HJ-NP-Z]{23,}')
    if not token_re.match(token):
        raise InvalidTokenError(token)


def to_bool(value):
    if isinstance(value, bool):
        return value
    elif isinstance(value, numbers.Number):
        return bool(value)
    else:
        raise Exception()


def to_numeric(value):
    if isinstance(value, bool):
        return float(value)
    elif isinstance(value, numbers.Number):
        return None if isnan(value) or isinf(value) else value
    elif is_string(value):
        log.warning(TYPE_WARN_MSG, value)
        return float(value)
    else:
        raise Exception()


def to_string(value):
    if isinstance(value, str):
        return value
    elif isinstance(value, bool):
        return str(value).lower()
    else:
        return str(value)


def to_location(value):
    if isinstance(value, dict):
        return {x: value[x] for x in ('lat', 'lng')}
    else:
        raise Exception()


def validate_variable_value(name, var_type, value):
    if value is None:
        return value

    convert = {
        BOOL_TYPE: to_bool,
        NUMERIC_TYPE: to_numeric,
        STRING_TYPE: to_string,
        LOCATION_TYPE: to_location
    }
    c = convert.get(var_type, None)
    if c is None:
        return None
    try:
        return c(value)
    except Exception:
        raise UnexpectedVariableValueTypeError('"{0}"={1}'.format(name, value))


def validate_config(cfg):
    if not isinstance(cfg, list):
        raise InvalidConfigError()

    for item in cfg:
        guard_against_invalid_variable_type(
            item.get('name', None),
            item.get('type', None)
        )
    return cfg


def guard_against_invalid_variable_type(name, var_type):
    if var_type not in SUPPORTED_VARIABLE_TYPES:
        raise UnexpectedVariableTypeError(name)


def utcnow():
    return datetime.utcnow().replace(tzinfo=UtcTzInfo()).isoformat()


def args_count(binding):
    # pylint: disable=E1101, W1505
    if inspect.getargspec is not None:
        args = inspect.getargspec(binding).args
    else:
        args = inspect.getfullargspec(binding).args

    return args.__len__()


def has_args(binding):
    if inspect.ismethod(binding):
        return args_count(binding) > 1
    else:
        return args_count(binding) > 0


def resolve_callable(binding, current=None):
    if has_args(binding):
        return binding(current)
    else:
        return binding()



# -*- coding: utf-8 -*-

from cloud4rpi import utils


class Device(object):
    def __init__(self, api):
        def on_command(cmd):
            self.__on_command(cmd)

        self.__api = api
        self.__api.on_command = on_command
        self.__variables = {}
        self.__diag = {}

    @staticmethod
    def __resolve_binding(binding, current=None, default=None):
        if hasattr(binding, 'read'):
            return binding.read()
        elif callable(binding):
            return utils.resolve_callable(binding, current)
        else:
            return default

    def __validate_payload(self, payload):
        result = {}
        for name, value in payload.items():
            variable = self.__variables.get(name, None)
            if not variable:
                continue
            t = variable.get('type', None)
            result[name] = utils.validate_variable_value(name, t, value)

        return result

    def __on_command(self, cmd):
        update = self.__apply_commands(cmd)
        if bool(update):
            self.__api.publish_data(update, data_type='cr')

    def __apply_commands(self, cmd):
        update = {}
        for varName, value in cmd.items():
            variable = self.__variables.get(varName, None)
            if not variable:
                continue
            # consider to use resolve binding here
            new_value = value
            handler = variable.get('bind', None)
            if callable(handler):
                new_value = handler(new_value)

            t = variable.get('type', None)
            new_value = utils.validate_variable_value(varName, t, new_value)
            variable['value'] = new_value

            update[varName] = new_value

        return update

    def declare(self, variables):
        for name, value in variables.items():
            utils.guard_against_invalid_variable_type(name,
                                                      value.get('type', None))
        self.__variables = variables

    def declare_diag(self, diag):
        self.__diag = diag

    def read_config(self):
        return [{'name': name, 'type': value['type']}
                for name, value in self.__variables.items()]

    def read_data(self):
        for name, varConfig in self.__variables.items():
            bind = varConfig.get('bind', None)
            if bind:
                curr = varConfig.get('value')
                result = self.__resolve_binding(bind, curr, curr)
                t = varConfig.get('type')
                new_val = utils.validate_variable_value(name, t, result)
                varConfig['value'] = new_val

        readings = {varName: varConfig.get('value')
                    for varName, varConfig in self.__variables.items()}

        return readings

    def read_diag(self):
        readings = {}
        for name, value in self.__diag.items():
            readings[name] = self.__resolve_binding(value, None, value)
        return readings

    def publish_config(self, cfg=None):
        if cfg is None:
            cfg = self.read_config()
        else:
            cfg = utils.validate_config(cfg)

        return self.__api.publish_config(cfg)

    def publish_data(self, data=None):
        if data is None:
            data = self.read_data()
        else:
            data = self.__validate_payload(data)

        return self.__api.publish_data(data)

    def publish_diag(self, diag=None):
        if diag is None:
            diag = self.read_diag()
        return self.__api.publish_diag(diag)



# -*- coding: utf-8 -*-
# These tests need a running mqtt broker instance

import os
import json
import unittest
from threading import Event
import paho.mqtt.client as mqtt

from cloud4rpi import MqttApi
from cloud4rpi.errors import InvalidTokenError


class MqttMessageProbe(object):
    def __init__(self, on_message):
        self.__client = mqtt.Client()
        self.__on_message = on_message
        self.__client.connect('localhost')
        self.last_message = None

        def on_msg(client, userdata, message):
            self.last_message = message
            on_message()

        self.__client.on_message = on_msg
        self.__client.subscribe('iot-hub/messages/#')
        self.__client.loop_start()

    def dispose(self):
        self.__client.loop_stop()
        self.__client.disconnect()


class MqttCommandPublisher(object):
    def __init__(self):
        self.__client = mqtt.Client()
        self.__client.connect('localhost')
        self.__client.loop_start()

    def dispose(self):
        self.__client.loop_stop()
        self.__client.disconnect()

    def publish_command(self, device_token, command):
        self.__client.publish('iot-hub/commands/{0}'.format(device_token),
                              payload=json.dumps(command))


class _TimeoutError(Exception):  # avoid from pylint redefined-builtin
    pass


class AsyncTestCase(unittest.TestCase):
    def setUp(self):
        super(AsyncTestCase, self).setUp()
        self._done = Event()
        self._done.clear()

    def tearDown(self):
        super(AsyncTestCase, self).tearDown()
        self._done.clear()

    def wait(self, timeout=None):
        if timeout is None:
            timeout = get_async_test_timeout()

        self._done.wait(timeout)

        if not self.is_done():
            raise _TimeoutError()

    def stop(self):
        self._done.set()

    def is_done(self):
        return self._done.isSet()


def get_async_test_timeout(default=5):
    try:
        return float(os.environ.get('ASYNC_TEST_TIMEOUT'))
    except (ValueError, TypeError):
        return default


@unittest.skip("The Real mqtt broker required")
class TestMqttApi(AsyncTestCase):
    def setUp(self):
        super(TestMqttApi, self).setUp()
        self.test_probe = MqttMessageProbe(on_message=self.stop)

    def tearDown(self):
        super(TestMqttApi, self).tearDown()
        self.test_probe.dispose()

    @staticmethod
    def create_api_client():
        client = MqttApi('4GPZFMVuacadesU21dBw47zJi', host='localhost')
        client.connect()
        return client

    def testCtorThrowsOnInvalidToken(self):
        with self.assertRaises(InvalidTokenError):
            MqttApi('invalid device token')

    def testPublishConfig(self):
        client = self.create_api_client()

        variables = [
            {'name': 'Temperature', 'type': 'numeric'},
            {'name': 'Cooler', 'type': 'bool'},
            {'name': 'TheAnswer', 'type': 'numeric'},
        ]
        client.publish_config(variables)

        self.wait()

        actual_msg = self.test_probe.last_message
        self.assertEqual('iot-hub/messages/4GPZFMVuacadesU21dBw47zJi',
                         actual_msg.topic)
        self.assertEqual('config', json.loads(actual_msg.payload)['type'])
        self.assertEqual(variables, json.loads(actual_msg.payload)['payload'])

    def testPublishData(self):
        client = self.create_api_client()

        data = {
            'Temperature': 36.6,
            'Cooler': True,
            'TheAnswer': 42
        }
        client.publish_data(data)

        self.wait()

        actual_msg = self.test_probe.last_message
        self.assertEqual('iot-hub/messages/4GPZFMVuacadesU21dBw47zJi',
                         actual_msg.topic)
        self.assertEqual('data', json.loads(actual_msg.payload)['type'])
        self.assertEqual(data, json.loads(actual_msg.payload)['payload'])

    def testPublishDiag(self):
        client = self.create_api_client()

        diag = {
            'IPAddress': '127.0.0.1',
            'Hostname': 'weather_station',
            'CPU Load': 99
        }
        client.publish_diag(diag)

        self.wait()

        actual_msg = self.test_probe.last_message
        self.assertEqual('iot-hub/messages/4GPZFMVuacadesU21dBw47zJi',
                         actual_msg.topic)
        self.assertEqual('system', json.loads(actual_msg.payload)['type'])
        self.assertEqual(diag, json.loads(actual_msg.payload)['payload'])

    def testOnCommand(self):
        def on_command(command):
            self.assertEqual({'Cooler': True}, command)
            self.stop()

        client = self.create_api_client()
        client.on_command = on_command

        commander = MqttCommandPublisher()
        commander.publish_command(
            '4GPZFMVuacadesU21dBw47zJi',
            {'Cooler': True}
        )

        self.wait()



# -*- coding: utf-8 -*-

import time
import logging
import json
import paho.mqtt.client as mqtt

from cloud4rpi import config
from cloud4rpi import utils
from cloud4rpi.errors import MqttConnectionError


KEEP_ALIVE_INTERVAL = 30  # sec
RETRY_INTERVAL = 5  # sec
CONNECT_RESULT_UNDEFINED = 255

log = logging.getLogger(config.loggerName)


def is_success(rc):
    return rc == mqtt.MQTT_ERR_SUCCESS


class MqttApi(object):
    def __init__(self,
                 device_token,
                 host=config.mqqtBrokerHost,
                 port=config.mqttBrokerPort,
                 tls_config=None):
        utils.guard_against_invalid_token(device_token)

        def noop_on_command(cmd):
            pass

        self.__device_token = device_token
        self.__client = mqtt.Client(device_token, clean_session=False)
        self.__host = host
        self.__port = port
        if isinstance(tls_config, dict):
            log.info('Configuring TLS')
            self.__client.tls_set(**tls_config)

        self.__qos = 1
        self.__connect_result = None

        self.on_command = noop_on_command
        self.__outgoing_messages = {}

    @property
    def commands_topic(self):
        return self.__format_topic('commands')

    @property
    def config_topic(self):
        return self.__format_topic('config')

    @property
    def data_topic(self):
        return self.__format_topic('data')

    @property
    def diag_topic(self):
        return self.__format_topic('diagnostics')

    def __format_topic(self, tail):
        return 'devices/{0}/{1}'.format(self.__device_token, tail)

    def connect(self):
        def on_connect(client, userdata, flags, rc):
            if not is_success(rc):
                log.error('Connection failed: %s', rc)
                raise MqttConnectionError(rc)

            log.info('Connected')
            self.__connect_result = rc
            self.__outgoing_messages = {}

            log.info('Subscribing %s with QoS %s',
                     self.commands_topic, str(self.__qos))
            self.__client.subscribe(self.commands_topic, qos=self.__qos)

        def on_message(client, userdata, msg):
            log.info('Command received %s: %s', msg.topic, msg.payload)
            if callable(self.on_command):
                payload = msg.payload.decode() \
                    if isinstance(msg.payload, bytes) else msg.payload
                self.on_command(json.loads(payload))

        def on_disconnect(client, userdata, rc):
            log.info('Disconnected with code: %s', rc)
            self.__on_disconnect(rc)

        def on_publish(client, packet, mid):
            info = self.__outgoing_messages.pop(mid, None)
            if info is not None:
                log.info('Published %s: %s', info['topic'], info['msg'])

        self.__client.on_connect = on_connect
        self.__client.on_message = on_message
        self.__client.on_disconnect = on_disconnect
        self.__client.on_publish = on_publish

        log.info('Connecting %s:%s', self.__host, self.__port)
        self.__client.connect(self.__host, self.__port,
                              keepalive=KEEP_ALIVE_INTERVAL)

        self.__connect_result = CONNECT_RESULT_UNDEFINED
        self.__client.loop_start()

        while self.__connect_result == CONNECT_RESULT_UNDEFINED:
            time.sleep(.01)

    def __on_disconnect(self, rc):
        if is_success(rc):
            return

        while True:
            log.info('Reconnecting')

            try:
                self.__client.reconnect()
                return
            except Exception as e:
                log.info('Reconnection failed: %s', str(e))

            time.sleep(RETRY_INTERVAL)

    def disconnect(self):
        self.__client.loop_stop()
        self.__client.disconnect()

    def publish_config(self, msg):
        self.__publish(self.config_topic, msg)

    def publish_data(self, msg, **kwargs):
        dt = kwargs.get('data_type')
        topic = '{0}/{1}'.format(self.data_topic, dt) if dt else self.data_topic

        self.__publish(topic, msg)

    def publish_diag(self, msg):
        self.__publish(self.diag_topic, msg)

    def __publish(self, topic, payload=None):
        if payload is None:
            return

        msg = {
            'ts': utils.utcnow(),
            'payload': payload,
        }

        (_, mid) = self.__client.publish(topic,
                                         qos=self.__qos,
                                         payload=json.dumps(msg))

        self.__outgoing_messages[mid] = {
            'topic': topic,
            'msg': msg,
        }



# -*- coding: utf-8 -*-

import subprocess

TYPE_WARN_MSG = 'WARNING! A string "%s" passed to a numeric variable. ' \
                'Change the variable type or the passed value.' \



class InvalidTokenError(Exception):
    pass


class InvalidConfigError(TypeError):
    pass


class UnexpectedVariableTypeError(TypeError):
    pass


class UnexpectedVariableValueTypeError(TypeError):
    pass


class MqttConnectionError(Exception):
    def __init__(self, code):
        super(MqttConnectionError, self).__init__()
        self.code = code


class NotSupportedError(Exception):
    pass


__messages = {
    KeyboardInterrupt: 'Interrupted',
    subprocess.CalledProcessError: 'Try run with sudo',
    InvalidTokenError:
        'Device token {0} is invalid. Please verify it.',
    InvalidConfigError:
        'Configuration is invalid. It must be an array.',
    UnexpectedVariableTypeError:
        ('Unexpected type for the "{0}" variable. '
         'It must be "bool", "numeric", "string" or "location".'),
    UnexpectedVariableValueTypeError:
        'Unexpected value type for variable: {0}'
}


def get_error_message(e):
    return __messages.get(type(e), 'Unexpected error: {0}').format(e.message)



# -*- coding: utf-8 -*-

import unittest
from mock import Mock
import cloud4rpi
from cloud4rpi.errors import InvalidConfigError
from cloud4rpi.errors import UnexpectedVariableTypeError
from cloud4rpi.errors import UnexpectedVariableValueTypeError


class ApiClientMock(object):
    def __init__(self):
        def noop_on_command(cmd):
            pass

        self.publish_config = Mock()
        self.publish_data = Mock()
        self.publish_diag = Mock()
        self.on_command = noop_on_command

    def assert_publish_data_called_with(self, expected):
        return self.publish_data.assert_called_with(expected, data_type='cr')

    def raise_on_command(self, cmd):
        self.on_command(cmd)


class MockSensor(object):
    def __init__(self, value=42):
        self.read = Mock(return_value=value)
        self.__innerValue__ = value

    def get_state(self):
        return self.__innerValue__

    def get_updated_state(self, value):
        self.__innerValue__ = value
        return self.__innerValue__

    def get_incremented_state(self, value):
        return self.__innerValue__ + value


class TestDevice(unittest.TestCase):
    def testDeclareVariables(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'CPUTemp': {
                'type': 'numeric',
                'bind': MockSensor()
            }
        })
        cfg = device.read_config()
        self.assertEqual(cfg, [{'name': 'CPUTemp', 'type': 'numeric'}])

    def testDeclareVariablesValidation(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        with self.assertRaises(UnexpectedVariableTypeError):
            device.declare({
                'CPUTemp': {
                    'type': 'number',
                    'bind': MockSensor()
                }
            })

    def testDeclareDiag(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare_diag({
            'IPAddress': '8.8.8.8',
            'Host': 'hostname',
        })
        diag = device.read_diag()
        self.assertEqual(diag, {'IPAddress': '8.8.8.8', 'Host': 'hostname'})

    def testReadConfig(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'SomeVar': {
                'type': 'string'
            }
        })
        device.declare({
            'CPUTemp': {
                'type': 'numeric',
                'bind': MockSensor()
            }
        })
        cfg = device.read_config()
        self.assertEqual(cfg, [{'name': 'CPUTemp', 'type': 'numeric'}])

    def testReadConfigIfNotDeclared(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        self.assertEqual(device.read_config(), [])

    def testReadVariables(self):
        handler = {}
        temperature_sensor = MockSensor(73)
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': handler
            },
            'Temperature': {
                'type': 'numeric',
                'value': True,
                'bind': temperature_sensor
            }
        })
        data = device.read_data()
        self.assertEqual(data, {
            'LEDOn': False,
            'Temperature': 73
        })

    def testReadVariablesDoesNotContainsEmptyVars(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        self.assertEqual(device.read_data(), {})

    def testReadVariablesFromClassMethod(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        sensor = MockSensor(10)

        device.declare({
            'MyParam': {
                'type': 'numeric',
                'bind': sensor.get_state
            },
        })
        data = device.read_data()
        self.assertEqual(data, {
            'MyParam': 10,
        })

    def testReadVariablesFromClassMethodWithCurrent(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        sensor = MockSensor(10)

        device.declare({
            'MyParam': {
                'type': 'numeric',
                'value': 1,
                'bind': sensor.get_incremented_state
            },
        })
        data = device.read_data()
        self.assertEqual(data, {
            'MyParam': 11,
        })

    def testReadDiag(self):
        temperature_sensor = MockSensor(73)
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare_diag({
            'CPUTemperature': temperature_sensor,
            'IPAddress': lambda x: '8.8.8.8',
            'OSName': lambda x: 'Linux',
            'Host': 'weather_station'
        })
        diag = device.read_diag()
        self.assertEqual(diag, {
            'CPUTemperature': 73,
            'IPAddress': '8.8.8.8',
            'OSName': 'Linux',
            'Host': 'weather_station'
        })

    def testPublishConfig(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        cfg = [
            {'name': 'CPUTemp', 'type': 'numeric'},
            {'name': 'Cooler', 'type': 'bool'}
        ]
        device.publish_config(cfg)
        api.publish_config.assert_called_with(cfg)

    def testReadBeforePublishConfig(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'CPUTemp': {
                'type': 'numeric',
                'bind': MockSensor()
            }
        })
        device.publish_config()
        cfg = [{'name': 'CPUTemp', 'type': 'numeric'}]
        api.publish_config.assert_called_with(cfg)

    def testPublishConfigFail_NotAnArray(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)

        cfg = {'name': 'CPUTemp', 'type': 'numeric'}
        with self.assertRaises(InvalidConfigError):
            device.publish_config(cfg)

        api.publish_config.assert_not_called()

    def testPublishConfigFail_UnexpectedVariableType(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)

        cfg = [{'name': 'CPUTemp', 'type': 'number'}]
        with self.assertRaises(UnexpectedVariableTypeError):
            device.publish_config(cfg)

        api.publish_config.assert_not_called()

    def testPublishDiag(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        diag = {
            'IPAddress': '8.8.8.8',
            'Host': 'hostname'
        }
        device.publish_diag(diag)
        api.publish_diag.assert_called_with(diag)

    def testReadBeforePublishDiag(self):
        temperature_sensor = MockSensor(24)
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare_diag({
            'CPUTemperature': temperature_sensor,
            'IPAddress': lambda x: '8.8.8.8',
        })
        device.publish_diag()
        diag = {'IPAddress': '8.8.8.8', 'CPUTemperature': 24}
        api.publish_diag.assert_called_with(diag)

    def testPublishVariablesOnlyData(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'Temperature': {
                'type': 'numeric'
            },
            'Cooler': {
                'type': 'bool',
            }
        })
        data = {
            'Temperature': 36.6,
            'Cooler': True,
            'TheAnswer': 42
        }
        device.publish_data(data)
        api.publish_data.assert_called_with({
            'Temperature': 36.6,
            'Cooler': True
        })

    def testPublishNotDeclaredVariables(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        data = {
            'Temperature': 36.6,
            'Cooler': True,
            'TheAnswer': 42
        }
        device.publish_data(data)
        api.publish_data.assert_called_with({})

    def testReadBeforePublishData(self):
        temperature_sensor = MockSensor(24)
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'Temperature': {
                'type': 'numeric',
                'value': True,
                'bind': temperature_sensor
            }
        })
        device.publish_data()
        data = {'Temperature': 24}
        api.publish_data.assert_called_with(data)

    def testDataReadValidation_Bool(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'CoolerOn': {
                'type': 'bool',
                'value': True,
                'bind': lambda x: 100
            }
        })
        device.publish_data()
        data = {'CoolerOn': True}
        api.publish_data.assert_called_with(data)

    def testDataReadValidation_Numeric(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'ReadyState': {
                'type': 'numeric',
                'value': True,
                'bind': lambda x: True
            }
        })
        device.publish_data()
        data = {'ReadyState': 1}
        api.publish_data.assert_called_with(data)

    def testDataReadValidation_String(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'ReadyState': {
                'type': 'string',
                'value': True,
                'bind': lambda x: True
            }
        })
        device.publish_data()
        data = {'ReadyState': 'true'}
        api.publish_data.assert_called_with(data)

    def testDataReadValidation_Location(self):
        api = ApiClientMock()
        device = cloud4rpi.Device(api)
        device.declare({
            'MyLocation': {
                'type': 'location',
                'value': True,
                'bind': lambda x: {'lat': 37.89, 'lng': 75.43}
            }
        })
        device.publish_data()
        data = {'MyLocation': {'lat': 37.89, 'lng': 75.43}}
        api.publish_data.assert_called_with(data)


class CommandHandling(unittest.TestCase):
    def setUp(self):
        super(CommandHandling, self).setUp()
        self.api = ApiClientMock()
        self.device = cloud4rpi.Device(self.api)

    def testCallsBoundFunction(self):
        handler = Mock(return_value=True)
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': handler
            }
        })
        self.api.raise_on_command({'LEDOn': True})
        handler.assert_called_with(True)

    def testCallsBoundFunctionWithAnArgument(self):
        sensor = MockSensor(0)
        self.device.declare({
            'Status': {
                'type': 'numeric',
                'value': 10,
                'bind': sensor.get_updated_state
            }
        })
        self.api.raise_on_command({'Status': 20})
        self.api.assert_publish_data_called_with({'Status': 20})

    def testBindIsNotCallableFunction(self):
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': 'this is not a function'
            }
        })
        expected = {'LEDOn': True}
        self.api.raise_on_command(expected)
        self.api.assert_publish_data_called_with(expected)

        data = self.device.read_data()
        self.assertEqual(data, expected)

    def testDirectUpdateVariableValue(self):
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
            }
        })
        expected = {'LEDOn': True}
        self.api.raise_on_command(expected)
        self.api.assert_publish_data_called_with(expected)
        data = self.device.read_data()
        self.assertEqual(data, expected)

    def testSkipUnknownVariable(self):
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': lambda x: x
            }
        })
        self.api.raise_on_command({'Other': True})
        self.api.publish_data.assert_not_called()

    def testAllowPublishNullValue(self):
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': lambda x: None
            }
        })
        self.api.raise_on_command({'LEDOn': True})
        self.api.assert_publish_data_called_with({'LEDOn': None})

    def testValidateCommandValueForBool(self):
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': lambda x: x
            }
        })
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.api.raise_on_command({'LEDOn': 'false'})

    def testValidateCommandValueStringToNumeric(self):
        self.device.declare({
            'Status': {
                'type': 'numeric',
                'value': 0,
                'bind': lambda x: x
            }
        })
        self.api.raise_on_command({'Status': '100'})
        self.api.assert_publish_data_called_with({'Status': 100})

    def testValidateCommandValueUnicodeToNumeric(self):
        self.device.declare({
            'Status': {
                'type': 'numeric',
                'value': 0,
                'bind': lambda x: x
            }
        })
        unicode_val = u'38.5'
        self.api.raise_on_command({'Status': unicode_val})
        self.api.assert_publish_data_called_with({'Status': 38.5})

    def testValidateCommandValueBoolToNumeric(self):
        self.device.declare({
            'Status': {
                'type': 'numeric',
                'value': 0,
                'bind': lambda x: x
            }
        })
        self.api.raise_on_command({'Status': True})
        self.api.assert_publish_data_called_with({'Status': 1})

    def testValidateCommandValueUnicodeToString(self):
        self.device.declare({
            'Percent': {
                'type': 'string',
                'value': 0,
                'bind': lambda x: x
            }
        })
        unicode_val = u'38.5%'
        self.api.raise_on_command({'Percent': unicode_val})
        self.api.assert_publish_data_called_with({'Percent': '38.5%'})

    def testPublishBackUpdatedVariableValues(self):
        sensor = MockSensor(36.6)
        self.device.declare({
            'LEDOn': {
                'type': 'bool',
                'value': False,
                'bind': lambda x: x
            },
            'Cooler': {
                'type': 'bool',
                'value': True,
                'bind': lambda x: x
            },
            'Status': {
                'type': 'numeric',
                'value': 0,
                'bind': lambda x: 42
            },
            'Temp': {
                'type': 'numeric',
                'value': 24.4,
                'bind': sensor
            }
        })
        self.api.raise_on_command({'LEDOn': True,
                                   'Cooler': False,
                                   'Status': 2,
                                   'Temp': 36.6})
        expected = {
            'Cooler': False,
            'Status': 42,
            'LEDOn': True,
            'Temp': 36.6
        }
        self.api.assert_publish_data_called_with(expected)

    def testPublishBackOnlyCommandVariables(self):
        self.device.declare({
            'Actuator': {
                'type': 'string',
                'value': 'to be updated and published',
                'bind': lambda x: x
            },
            'Sensor': {
                'type': 'string',
                'value': None,
                'bind': 'do not updated by a command'
            },
        })
        self.api.raise_on_command({'Actuator': 'ON'})
        self.api.assert_publish_data_called_with({'Actuator': 'ON'})


class PayloadValidation(unittest.TestCase):
    def setUp(self):
        super(PayloadValidation, self).setUp()
        self.api = ApiClientMock()
        self.device = cloud4rpi.Device(self.api)

    def testNumeric(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': 36.3})
        self.api.publish_data.assert_called_with({'Temp': 36.3})

    def testNumericAsNull(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': None})
        self.api.publish_data.assert_called_with({'Temp': None})

    def testNumericAsInt(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': 36})
        self.api.publish_data.assert_called_with({'Temp': 36})

    def testNumericAsFloat(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': 36.6})
        self.api.publish_data.assert_called_with({'Temp': 36.6})

    def testNumericAsString(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': "36.6"})
        self.api.publish_data.assert_called_with({'Temp': 36.6})

    def testNumericAsBool(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': True})
        self.api.publish_data.assert_called_with({'Temp': 1.0})

    def testNumericAsNaN(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': float('NaN')})
        self.api.publish_data.assert_called_with({'Temp': None})

    def testNumericAsPositiveInfinity(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': float('Inf')})
        self.api.publish_data.assert_called_with({'Temp': None})

    def testNumericAsNegativeInfinity(self):
        self.device.declare({'Temp': {'type': 'numeric'}})
        self.device.publish_data({'Temp': -float('Inf')})
        self.api.publish_data.assert_called_with({'Temp': None})

    def testBool(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': True})
        self.api.publish_data.assert_called_with({'PowerOn': True})

    def testBoolAsNull(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': None})
        self.api.publish_data.assert_called_with({'PowerOn': None})

    def testBoolAsString(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'PowerOn': "True"})

    def testBoolAsPositiveNumber(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': 24.1})
        self.api.publish_data.assert_called_with({'PowerOn': True})

    def testBoolAsNegativeNumber(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': -10.1})
        self.api.publish_data.assert_called_with({'PowerOn': True})

    def testBoolAsZeroNumber(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': 0})
        self.api.publish_data.assert_called_with({'PowerOn': False})

    def testBoolAsNaN(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': float('NaN')})
        self.api.publish_data.assert_called_with({'PowerOn': True})

    def testBoolAsPositiveInfinity(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': float('Inf')})
        self.api.publish_data.assert_called_with({'PowerOn': True})

    def testBoolAsNegativeInfinity(self):
        self.device.declare({'PowerOn': {'type': 'bool'}})
        self.device.publish_data({'PowerOn': -float('Inf')})
        self.api.publish_data.assert_called_with({'PowerOn': True})

    def testString(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': '100'})
        self.api.publish_data.assert_called_with({'Status': '100'})

    def testStringAsNull(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': None})
        self.api.publish_data.assert_called_with({'Status': None})

    def testStringAsNumeric(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': 100.100})
        self.api.publish_data.assert_called_with({'Status': '100.1'})

    def testStringAsNaN(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': float('NaN')})
        self.api.publish_data.assert_called_with({'Status': 'nan'})

    def testStringAsPositiveInfinity(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': float('Inf')})
        self.api.publish_data.assert_called_with({'Status': 'inf'})

    def testStringAsNegativeInfinity(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': -float('Inf')})
        self.api.publish_data.assert_called_with({'Status': '-inf'})

    def testStringAsInt(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': 100})
        self.api.publish_data.assert_called_with({'Status': '100'})

    def testStringAsBool(self):
        self.device.declare({'Status': {'type': 'string'}})
        self.device.publish_data({'Status': True})
        self.api.publish_data.assert_called_with({'Status': 'true'})

    def testLocation(self):
        location = {'lat': 37.89, 'lng': 75.43}
        self.device.declare({'Pos': {'type': 'location'}})
        self.device.publish_data({'Pos': location})
        self.api.publish_data.assert_called_with({'Pos': location})

    def testLocation_Filtering(self):
        obj = {'some': 'foo', 'lng': 75.43, 'lat': 37.89, 'other': 42}
        self.device.declare({'Pos': {'type': 'location'}})
        self.device.publish_data({'Pos': obj})
        location = {'lat': 37.89, 'lng': 75.43}
        self.api.publish_data.assert_called_with({'Pos': location})

    def testLocationAsNull(self):
        self.device.declare({'Pos': {'type': 'location'}})
        self.device.publish_data({'Pos': None})
        self.api.publish_data.assert_called_with({'Pos': None})

    def testLocationAsNaN(self):
        self.device.declare({'Pos': {'type': 'location'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'Pos': float('NaN')})

    def testLocationAsInfinity(self):
        self.device.declare({'Pos': {'type': 'location'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'Pos': float('Inf')})

    def testLocationAsEmptyObject(self):
        self.device.declare({'Pos': {'type': 'location'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'Pos': {}})

    def testLocationWithIncorrectFields(self):
        location = {'Latitude': 37.89, 'LNG': 75.43}
        self.device.declare({'Pos': {'type': 'location'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'Pos': location})

    def testLocationWithoutLatitude(self):
        location = {'lng': 75.43}
        self.device.declare({'Pos': {'type': 'location'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'Pos': location})

    def testLocationWithoutLongitude(self):
        location = {'lat': 37.89}
        self.device.declare({'Pos': {'type': 'location'}})
        with self.assertRaises(UnexpectedVariableValueTypeError):
            self.device.publish_data({'Pos': location})



from setuptools import setup

description = ''

try:
    import pypandoc  # pylint: disable=F0401

    description = pypandoc.convert('README.md', 'rst')
except Exception:
    pass

setup(name='cloud4rpi',
      version='1.0.0',
      description='Cloud4RPi client library',
      long_description=description,
      url='https://github.com/cloud4rpi/cloud4rpi',
      author='Cloud4RPi team',
      author_email='team@cloud4rpi.io',
      license='MIT',
      packages=['cloud4rpi'],
      install_requires=[
          'paho-mqtt'
      ])
